#include <glib.h>
#include <dbus/dbus-glib.h>
#include <stdlib.h>
#include <unistd.h> 
#include <string.h>

#include "common-defs.h"

typedef enum {
	E_SIGNAL_MESSAGE,
	E_SIGNAL_MYSIGNAL,
	E_SIGNAL_COUNT
} ValueSignalNumber;


typedef struct {
	GObject parent;
	gchar *mymessage;
	gchar *mysignal;
} ValueObject;

typedef struct {
	GObjectClass parent;
	gint thresholdMin;
	gint thresholdMax;
	guint signals[E_SIGNAL_COUNT];
} ValueObjectClass;

GType value_object_get_type(void);

static gint sequence_number = 0;

#define VALUE_TYPE_OBJECT              (value_object_get_type())
#define VALUE_OBJECT(object) (G_TYPE_CHECK_INSTANCE_CAST ((object), VALUE_TYPE_OBJECT, ValueObject))

#define VALUE_OBJECT_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), VALUE_TYPE_OBJECT, ValueObjectClass))
#define VALUE_IS_OBJECT(object) (G_TYPE_CHECK_INSTANCE_TYPE ((object),VALUE_TYPE_OBJECT))

#define VALUE_IS_OBJECT_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), VALUE_TYPE_OBJECT))
#define VALUE_OBJECT_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), VALUE_TYPE_OBJECT, ValueObjectClass))

G_DEFINE_TYPE(ValueObject, value_object, G_TYPE_OBJECT)

gboolean value_object_setmymessage(ValueObject *obj, gchar *value_in, GError** error);
gboolean value_object_getmymessage(ValueObject *obj, gchar *value_out, GError** error);

#include "value-server-stub.h"

#ifdef NO_DAEMON
#define dbg(fmtstr, args...) \
  (g_print(PROGNAME ":%s: " fmtstr "\n", __func__, ##args))
#else
#define dbg(dummy...)
#endif

static void value_object_init(ValueObject* obj) {
	dbg("Called");
	g_assert(obj != NULL);
	obj->mymessage = "NULL";
	obj->mysignal = "NULL";
}

static void value_object_class_init(ValueObjectClass* klass) {
	
	const gchar* signalNames[E_SIGNAL_COUNT] = {	SIGNAL_STRING_MESSAGE, SIGNAL_STRING_MYSIGNAL };
	
	dbg("Called");
	g_assert(klass != NULL);
	
	klass->thresholdMin = -100;
	klass->thresholdMax = 100;

	dbg("Creating signals");

	int i;
	
	for(i=0; i < E_SIGNAL_COUNT; i++)
	{
		guint signalId;
	
		signalId =g_signal_new(signalNames[i],G_OBJECT_CLASS_TYPE(klass),G_SIGNAL_RUN_LAST,0,NULL, NULL,g_cclosure_marshal_VOID__STRING,G_TYPE_NONE, 1,G_TYPE_STRING);
	
		klass->signals[i] = signalId;
	}

	dbg("Binding to GLib/D-Bus");

	dbus_g_object_type_install_info(VALUE_TYPE_OBJECT,&dbus_glib_value_object_object_info);

	dbg("Done");
}

static void value_object_emitSignal(ValueObject* obj,ValueSignalNumber num,const gchar* message) 
{
	ValueObjectClass* klass = VALUE_OBJECT_GET_CLASS(obj);
	g_assert((num < E_SIGNAL_COUNT) && (num >= 0));
	dbg("Emitting signal id %d, with message '%s'", num, message);

	g_signal_emit(obj,klass->signals[num],0,message);

}

static void handleError(const char* msg, const char* reason,gboolean fatal) 
{
	g_printerr(PROGNAME ": ERROR: %s (%s)\n", msg, reason);
	if (fatal) 
	{
		exit(EXIT_FAILURE);
	}
}

gboolean value_object_setmymessage(ValueObject* obj, gchar *valueIn,GError** error) 
{
	dbg("Incoming Sequence Number =%s", valueIn);
	g_assert(obj != NULL);
	sequence_number++;
	dbg("Server Sequence Number =%d",sequence_number);
	return TRUE;
}

gboolean value_object_getmymessage(ValueObject* obj, gchar *valueOut,GError** error) 
{
	dbg("Called (internal mymessage is %s)", obj->mymessage);
	g_assert(obj != NULL);
	g_assert(valueOut != NULL);
	valueOut = obj->mymessage;

	sleep(10);
	
	return TRUE;
}	


int main(int argc, char** argv) {
	DBusGConnection* bus = NULL;
	DBusGProxy* busProxy = NULL;
	ValueObject* valueObj = NULL;
	GMainLoop* mainloop = NULL;
	guint result;
	GError* error = NULL;

	g_type_init();

	mainloop = g_main_loop_new(NULL, FALSE);
	if (mainloop == NULL) 
	{
		handleError("Couldn't create GMainLoop", "Unknown(OOM?)", TRUE);
	}

	g_print(PROGNAME ":main Connecting to the System D-Bus.\n");
	bus = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);
	if (error != NULL) 
	{
		handleError("Couldn't connect to system bus", error->message, TRUE);
	}

	g_print(PROGNAME ":main Registering the well-known name (%s)\n",VALUE_SERVICE_NAME);

	busProxy = dbus_g_proxy_new_for_name(bus,DBUS_SERVICE_DBUS,DBUS_PATH_DBUS,DBUS_INTERFACE_DBUS);
	
	if (busProxy == NULL) 
	{
		handleError("Failed to get a proxy for D-Bus","Unknown(dbus_g_proxy_new_for_name)", TRUE);
	}

	if (!dbus_g_proxy_call(busProxy,"RequestName",&error,G_TYPE_STRING, VALUE_SERVICE_NAME,G_TYPE_UINT,0,G_TYPE_INVALID,G_TYPE_UINT,&result,G_TYPE_INVALID)) 
	{
		handleError("D-Bus.RequestName RPC failed", error->message,TRUE);
	}

	g_print(PROGNAME ":main RequestName returned %d.\n", result);

	if (result != 1) 
	{
		handleError("Failed to get the primary well-known name.","RequestName result != 1", TRUE);
	}

	g_print(PROGNAME ":main Registering signal handler signatures.\n");

	g_print(PROGNAME ":main Creating one Value object.\n");

	valueObj = g_object_new(VALUE_TYPE_OBJECT, NULL);
	if (valueObj == NULL) 
	{
		handleError("Failed to create one Value instance.","Unknown(OOM?)", TRUE);
	}

	g_print(PROGNAME ":Flushing the connection to remove earlier messages\n");
	dbus_g_connection_flush(bus);
	
	g_print(PROGNAME ":main Registering it on the D-Bus.\n");
	dbus_g_connection_register_g_object(bus,VALUE_SERVICE_OBJECT_PATH,G_OBJECT(valueObj));

	g_print(PROGNAME ":main Ready to serve requests (daemonizing).\n");

	#ifndef NO_DAEMON

	if (daemon(0, 0) != 0) 
	{
		g_error(PROGNAME ": Failed to daemonize.\n");
	}
	#else
	g_print(PROGNAME": Not daemonizing (built with NO_DAEMON-build define)\n");
	#endif

	g_main_loop_run(mainloop);

	return EXIT_FAILURE;
}
